/* bsgs.c -- (C) Geoffrey Reynolds, April 2006.

   Implementation of a baby step giant step algorithm for finding all
   n in the range nmin <= n <= nmax satisfying b^n=d_i (mod p) where b
   and each d_i are relatively prime to p.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
*/

#include <assert.h>
#include <inttypes.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "sr1sieve.h"
#include "arithmetic.h"
#include "bitmap.h"
#include "hashtable.h"

/*
  Giant step baby step algorithm, from Wikipaedea:

  input: A cyclic group G of order n, having a generator a and an element b.

  Output: A value x satisfying a^x = b (mod n).

  1. m <-- ceiling(sqrt(n))
  2. For all j where 0 <= j < m:
     1. Compute a^j mod n and store the pair (j, a^j) in a table.
  3. Compute a^(-m).
  4. g <-- b.
  5. For i = 0 to (m - 1):
     1. Check if g is the second component (a^j) of any pair in the table.
     2. If so, return im + j.
     3. If not, g <-- ga^(-m) mod n.
*/


/* Pre-compute the largest power bdQ=b^(Q/dQ) < p_min, where dQ divides Q.
   This allows b^Q (mod p) to be computed later as bdQ^dQ (mod p).
*/
#define PRE_POWER 1

static uint32_t m;     /* Number of baby steps */
static uint32_t M;     /* Number of giant steps */
static uint32_t sieve_low;

typedef struct {uint32_t m; uint32_t M;} steps_t;
static steps_t *steps;

#if PRE_POWER
static uint32_t dQ;           /* Minimum dQ|Q such that b^(Q/dQ)<p_min */
static uint64_t bdQ;          /* b^(Q/dQ) */
#endif

#ifdef HAVE_vector_sse2
static uint32_t baby_steps_sse2_2(uint64_t p)
{
  uint64_t b, bj0;
  uint32_t j;

  /* b <- base^Q (mod p) */
#if PRE_POWER
  b = powmod64(bdQ,dQ,p);
#elif (BASE == 0)
  b = powmod64(b_term,subseq_Q,p);
#else
  b = powmod64(BASE,subseq_Q,p);
#endif

  bj0 = powmod64(b,sieve_low,p);
  BJ64[0] = bj0;
  BJ64[1] = mulmod64(bj0,b,p);

  /**/ /* Unnecessary in the unlikely case m <= 2 */
  b = sqrmod64(b,p);
  vec_mulmod64_initb_sse2(&b);
  vec2_mulmod64_sse2(BJ64,BJ64+2,m-2);
  vec_mulmod64_finib_sse2();
  /**/
  BJ64[m] = bj0;

  clear_hashtable();
  j = 0;
  do { post_insert64(j); }
  while (BJ64[++j] != bj0);

  return (j < m) ? j : 0;
}

static uint32_t baby_steps_sse2_4(uint64_t p)
{
  uint64_t b, bj0;
  uint32_t j;

  /* b <- base^Q (mod p) */
#if PRE_POWER
  b = powmod64(bdQ,dQ,p);
#elif (BASE == 0)
  b = powmod64(b_term,subseq_Q,p);
#else
  b = powmod64(BASE,subseq_Q,p);
#endif

  bj0 = powmod64(b,sieve_low,p);
  BJ64[0] = bj0;
  PRE2_MULMOD64_INIT(b);
  for (j = 0; j < 3; j++)
    BJ64[j+1] = PRE2_MULMOD64(BJ64[j],b,p);
  PRE2_MULMOD64_FINI();

  /**/ /* Unnecessary in the unlikely case m <= 4 */
  b = powmod64(b,4,p);
  vec_mulmod64_initb_sse2(&b);
  vec4_mulmod64_sse2(BJ64,BJ64+4,m-4);
  vec_mulmod64_finib_sse2();
  /**/
  BJ64[m] = bj0;

  clear_hashtable();
  j = 0;
  do { post_insert64(j); }
  while (BJ64[++j] != bj0);

  return (j < m) ? j : 0;
}

static uint32_t baby_steps_sse2_8(uint64_t p)
{
  uint64_t b, bj0;
  uint32_t j;

  /* b <- base^Q (mod p) */
#if PRE_POWER
  b = powmod64(bdQ,dQ,p);
#elif (BASE == 0)
  b = powmod64(b_term,subseq_Q,p);
#else
  b = powmod64(BASE,subseq_Q,p);
#endif

  bj0 = powmod64(b,sieve_low,p);
  BJ64[0] = bj0;
  BJ64[1] = mulmod64(bj0,b,p);
  b = sqrmod64(b,p);
  vec_mulmod64_initb_sse2(&b);
  vec2_mulmod64_sse2(BJ64,BJ64+2,6);
  vec_mulmod64_finib_sse2();

  /**/ /* Unnecessary in the unlikely case m <= 8 */
  b = powmod64(b,4,p);
  vec_mulmod64_initb_sse2(&b);
  vec8_mulmod64_sse2(BJ64,BJ64+8,m-8);
  vec_mulmod64_finib_sse2();
  /**/
  BJ64[m] = bj0;

  clear_hashtable();
  j = 0;
  do { post_insert64(j); }
  while (BJ64[++j] != bj0);

  return (j < m) ? j : 0;
}

static uint32_t baby_steps_sse2_16(uint64_t p)
{
  uint64_t b, bj0;
  uint32_t j;

  /* b <- base^Q (mod p) */
#if PRE_POWER
  b = powmod64(bdQ,dQ,p);
#elif (BASE == 0)
  b = powmod64(b_term,subseq_Q,p);
#else
  b = powmod64(BASE,subseq_Q,p);
#endif

  bj0 = powmod64(b,sieve_low,p);
  BJ64[0] = bj0;
  BJ64[1] = mulmod64(bj0,b,p);
  b = sqrmod64(b,p);
  vec_mulmod64_initb_sse2(&b);
  vec2_mulmod64_sse2(BJ64,BJ64+2,14);
  vec_mulmod64_finib_sse2();

  /**/ /* Unnecessary in the unlikely case m <= 16 */
  b = powmod64(b,8,p);
  vec_mulmod64_initb_sse2(&b);
  vec16_mulmod64_sse2(BJ64,BJ64+16,m-16);
  vec_mulmod64_finib_sse2();
  /**/
  BJ64[m] = bj0;

  clear_hashtable();
  j = 0;
  do { post_insert64(j); }
  while (BJ64[++j] != bj0);

  return (j < m) ? j : 0;
}
#endif /* HAVE_vector_sse2 */

#ifdef HAVE_vector
static uint32_t baby_steps_2(uint64_t p)
{
  uint64_t b, bj0;
  uint32_t j;

  /* b <- base^Q (mod p) */
#if PRE_POWER
  b = powmod64(bdQ,dQ,p);
#else
  b = powmod64(b_term,subseq_Q,p);
#endif

  bj0 = powmod64(b,sieve_low,p);
  BJ64[0] = bj0;
  BJ64[1] = mulmod64(bj0,b,p);

  b = sqrmod64(b,p);
#ifdef __i386__
  vec_mulmod64_initb_i386(&b);
  vec2_mulmod64_i386(BJ64,BJ64+2,m-2);
  vec_mulmod64_finib_i386();
#else
  PRE2_MULMOD64_INIT(b);
  for (j = 0; j+2 < m; j += 2)
    VEC2_MULMOD64_NEXT(BJ64+j,b,p);
  PRE2_MULMOD64_FINI();
#endif
  BJ64[m] = bj0;

  clear_hashtable();
  j = 0;
  do { post_insert64(j); }
  while (BJ64[++j] != bj0);

  return (j < m) ? j : 0;
}

static uint32_t baby_steps_4(uint64_t p)
{
  uint64_t b, bj0;
  uint32_t j;

  /* b <- base^Q (mod p) */
#if PRE_POWER
  b = powmod64(bdQ,dQ,p);
#else
  b = powmod64(b_term,subseq_Q,p);
#endif

  bj0 = powmod64(b,sieve_low,p);
  BJ64[0] = bj0;
  PRE2_MULMOD64_INIT(b);
  for (j = 0; j < 3; j++)
    BJ64[j+1] = PRE2_MULMOD64(BJ64[j],b,p);
  PRE2_MULMOD64_FINI();

  b = powmod64(b,4,p);
#ifdef __i386__
  vec_mulmod64_initb_i386(&b);
  vec4_mulmod64_i386(BJ64,BJ64+4,m-4);
  vec_mulmod64_finib_i386();
#else
  PRE2_MULMOD64_INIT(b);
  for (j = 0; j+4 < m; j += 4)
    VEC4_MULMOD64_NEXT(BJ64+j,b,p);
  PRE2_MULMOD64_FINI();
#endif
  BJ64[m] = bj0;

  clear_hashtable();
  j = 0;
  do { post_insert64(j); }
  while (BJ64[++j] != bj0);

  return (j < m) ? j : 0;
}

static uint32_t baby_steps_8(uint64_t p)
{
  uint64_t b, bj0;
  uint32_t j;

  /* b <- base^Q (mod p) */
#if PRE_POWER
  b = powmod64(bdQ,dQ,p);
#else
  b = powmod64(b_term,subseq_Q,p);
#endif

  bj0 = powmod64(b,sieve_low,p);
  BJ64[0] = bj0;
  PRE2_MULMOD64_INIT(b);
  for (j = 0; j < 7; j++)
    BJ64[j+1] = PRE2_MULMOD64(BJ64[j],b,p);
  PRE2_MULMOD64_FINI();

  b = powmod64(b,8,p);
#ifdef __i386__
  vec_mulmod64_initb_i386(&b);
  vec8_mulmod64_i386(BJ64,BJ64+8,m-8);
  vec_mulmod64_finib_i386();
#else
  PRE2_MULMOD64_INIT(b);
  for (j = 0; j+8 < m; j += 8)
    VEC8_MULMOD64_NEXT(BJ64+j,b,p);
  PRE2_MULMOD64_FINI();
#endif
  BJ64[m] = bj0;

  clear_hashtable();
  j = 0;
  do { post_insert64(j); }
  while (BJ64[++j] != bj0);

  return (j < m) ? j : 0;
}
#endif /* HAVE_vector */

static uint32_t baby_steps_1(uint64_t p)
{
  uint64_t b, bj, bj0;
  uint32_t j;

  /* b <- base^Q (mod p) */
#if PRE_POWER
  b = powmod64(bdQ,dQ,p);
#else
  b = powmod64(b_term,subseq_Q,p);
#endif

  bj0 = powmod64(b,sieve_low,p);
  clear_hashtable();
  PRE2_MULMOD64_INIT(b);
  for (j = 0, bj = bj0; j < m; j++)
  {
    insert64(j, bj);
    bj = PRE2_MULMOD64(bj,b,p);
    if (bj == bj0) /* Unlikely */
    {
      PRE2_MULMOD64_FINI();
      return j+1;
    }
  }
  PRE2_MULMOD64_FINI();

  return 0;
}

static uint64_t bQ;           /* 1/b^Q (mod p) */
static const uint16_t *C16;   /* List of candidate subsequences for BSGS */
static uint64_t *D64;         /* D64[j] holds -c/(k*b^(im+d)) (mod p) for
                                 subsequence C16[j] at giant step i. */

#ifdef HAVE_vector_sse2
static void giant_steps_sse2_2(uint64_t p, uint32_t cc)
{
  uint64_t b;
  uint32_t i, j, k;

  assert (M > 1);
  assert (cc > 0);

  b = powmod64(bQ,m,p); /* b <- 1/b^Qm (mod p) */
  vec_mulmod64_initb_sse2(&b);
  i = 1;
  do
  {
    vec2_mulmod64_sse2(D64,D64,cc);
    for (k = 0; k < cc; k++)
      if ((j = lookup64(D64[k])) != HASH_NOT_FOUND)
        eliminate_term(C16[k],sieve_low+i*m+j,p);
  } while (++i < M);
  vec_mulmod64_finib_sse2();
}

static void giant_steps_sse2_4(uint64_t p, uint32_t cc)
{
  uint64_t b;
  uint32_t i, j, k;

  assert (M > 1);
  assert (cc > 0);

  b = powmod64(bQ,m,p); /* b <- 1/b^Qm (mod p) */
  vec_mulmod64_initb_sse2(&b);
  i = 1;
  do
  {
    vec4_mulmod64_sse2(D64,D64,cc);
    for (k = 0; k < cc; k++)
      if ((j = lookup64(D64[k])) != HASH_NOT_FOUND)
        eliminate_term(C16[k],sieve_low+i*m+j,p);
  } while (++i < M);
  vec_mulmod64_finib_sse2();
}

static void giant_steps_sse2_8(uint64_t p, uint32_t cc)
{
  uint64_t b;
  uint32_t i, j, k;

  assert (M > 1);
  assert (cc > 0);

  b = powmod64(bQ,m,p); /* b <- 1/b^Qm (mod p) */
  vec_mulmod64_initb_sse2(&b);
  i = 1;
  do
  {
    vec8_mulmod64_sse2(D64,D64,cc);
    for (k = 0; k < cc; k++)
      if ((j = lookup64(D64[k])) != HASH_NOT_FOUND)
        eliminate_term(C16[k],sieve_low+i*m+j,p);
  } while (++i < M);
  vec_mulmod64_finib_sse2();
}

static void giant_steps_sse2_16(uint64_t p, uint32_t cc)
{
  uint64_t b;
  uint32_t i, j, k;

  assert (M > 1);
  assert (cc > 0);

  b = powmod64(bQ,m,p); /* b <- 1/b^Qm (mod p) */
  vec_mulmod64_initb_sse2(&b);
  i = 1;
  do
  {
    vec16_mulmod64_sse2(D64,D64,cc);
    for (k = 0; k < cc; k++)
      if ((j = lookup64(D64[k])) != HASH_NOT_FOUND)
        eliminate_term(C16[k],sieve_low+i*m+j,p);
  } while (++i < M);
  vec_mulmod64_finib_sse2();
}
#endif /* HAVE_vector_sse2 */

#ifdef HAVE_vector
static void giant_steps_2(uint64_t p, uint32_t cc)
{
  uint64_t b;
  uint32_t i, j, k;

  assert (M > 1);
  assert (cc > 0);

  b = powmod64(bQ,m,p); /* b <- 1/b^Qm (mod p) */
#ifdef __i386__
  vec_mulmod64_initb_i386(&b);
  i = 1;
  do
  {
    vec2_mulmod64_i386(D64,D64,cc);
    for (k = 0; k < cc; k++)
      if ((j = lookup64(D64[k])) != HASH_NOT_FOUND)
        eliminate_term(C16[k],sieve_low+i*m+j,p);
  } while (++i < M);
  vec_mulmod64_finib_i386();
#else
  PRE2_MULMOD64_INIT(b);
  i = 1;
  do
  {
    for (k = 0; k < cc; k += 2)
      VEC2_MULMOD64(D64+k,b,p);
    for (k = 0; k < cc; k++)
      if ((j = lookup64(D64[k])) != HASH_NOT_FOUND)
        eliminate_term(C16[k],sieve_low+i*m+j,p);
  } while (++i < M);
  PRE2_MULMOD64_FINI();
#endif
}

static void giant_steps_4(uint64_t p, uint32_t cc)
{
  uint64_t b;
  uint32_t i, j, k;

  assert (M > 1);
  assert (cc > 0);

#ifndef __i386__
  if (cc <= 2)
  {
    giant_steps_2(p,cc);
    return;
  }
#endif

  b = powmod64(bQ,m,p); /* b <- 1/b^Qm (mod p) */
#ifdef __i386__
  vec_mulmod64_initb_i386(&b);
  i = 1;
  do
  {
    vec4_mulmod64_i386(D64,D64,cc);
    for (k = 0; k < cc; k++)
      if ((j = lookup64(D64[k])) != HASH_NOT_FOUND)
        eliminate_term(C16[k],sieve_low+i*m+j,p);
  } while (++i < M);
  vec_mulmod64_finib_i386();
#else
  PRE2_MULMOD64_INIT(b);
  i = 1;
  do
  {
    for (k = 0; k < cc; k += 4)
      VEC4_MULMOD64(D64+k,b,p);
    for (k = 0; k < cc; k++)
      if ((j = lookup64(D64[k])) != HASH_NOT_FOUND)
        eliminate_term(C16[k],sieve_low+i*m+j,p);
  } while (++i < M);
  PRE2_MULMOD64_FINI();
#endif
}

static void giant_steps_8(uint64_t p, uint32_t cc)
{
  uint64_t b;
  uint32_t i, j, k;

  assert (M > 1);
  assert (cc > 0);

#ifndef __i386__
  if (cc <= 4)
  {
    giant_steps_4(p,cc);
    return;
  }
#endif

  b = powmod64(bQ,m,p); /* b <- 1/b^Qm (mod p) */
#ifdef __i386__
  vec_mulmod64_initb_i386(&b);
  i = 1;
  do
  {
    vec8_mulmod64_i386(D64,D64,cc);
    for (k = 0; k < cc; k++)
      if ((j = lookup64(D64[k])) != HASH_NOT_FOUND)
        eliminate_term(C16[k],sieve_low+i*m+j,p);
  } while (++i < M);
  vec_mulmod64_finib_i386();
#else
  PRE2_MULMOD64_INIT(b);
  i = 1;
  do
  {
    for (k = 0; k < cc; k += 8)
      VEC8_MULMOD64(D64+k,b,p);
    for (k = 0; k < cc; k++)
      if ((j = lookup64(D64[k])) != HASH_NOT_FOUND)
        eliminate_term(C16[k],sieve_low+i*m+j,p);
  } while (++i < M);
  PRE2_MULMOD64_FINI();
#endif
}
#endif /* HAVE_vector */

static void giant_steps_1(uint64_t p, uint32_t cc)
{
  uint64_t b;
  uint32_t i, j, k;

  assert (M > 1);
  assert (cc > 0);

  b = powmod64(bQ,m,p); /* b <- 1/b^Qm (mod p) */
  PRE2_MULMOD64_INIT(b);
  i = 1;
  do
  {
    for (k = 0; k < cc; k++)
    {
      D64[k] = PRE2_MULMOD64(D64[k],b,p);
      if ((j = lookup64(D64[k])) != HASH_NOT_FOUND)
        eliminate_term(C16[k],sieve_low+i*m+j,p);
    }
  } while (++i < M);
  PRE2_MULMOD64_FINI();
}


/* Assign BJ64[i] = 1/b^i (mod p) for each i in the ladder.
   Return 1/b^Q (mod p).
 */
static uint64_t climb_ladder(const uint16_t *ladder, uint64_t inv, uint64_t p)
{
  uint32_t i;

  BJ64[0] = 1;
  BJ64[1] = inv;
  BJ64[2] = sqrmod64(inv,p);
  for (i = 2; *ladder > 0; i += *ladder++)
    BJ64[i+*ladder] = mulmod64(BJ64[i],BJ64[*ladder],p);

  return BJ64[subseq_Q];
}

/* This function builds the list C16[] of subsequences (k*b^d)*(b^Q)^m+c for
   which p may be a factor (-ckb^d is a quadratic/cubic/quartic/quintic
   residue with respect to p) and initialises the table D64[] with the
   values -c/(k*b^d) (mod p). As a side effect, bQ is set to the value 1/b^Q
   (mod p) for use later in bsgs64(), and m,M are set to the best choice of
   baby,giant steps. Returns the number of subsequences listed in C16[].
*/
static uint32_t attribute ((noinline)) setup64(uint64_t p, int parity)
{
  uint64_t inv, p_s, X[POWER_RESIDUE_LCM+1];
  const uint16_t *list;
  uint32_t r, h, j;

  /* inv <-- 1/b (mod p) */
  inv = (b_term == 2) ? (p+1)/2 : invmod32_64(b_term,p);

  int16_t s = div_shift[(p/2)%(POWER_RESIDUE_LCM/2)];
  if (s == 0)
  {
    /* p = 1 (mod 2) is all we know, check for quadratic residues only.
     */

    /* Precompute 1/b^d (mod p) for 0 <= d <= Q, as necessary. */
    bQ = climb_ladder(sc_ladders[parity+1][0][0],inv,p);

    /* For each subsequence (k*b^d)*(b^Q)^(n/Q)+c, compute
       -c/(k*b^d) (mod p) and add the subsequence to the bsgs list.
    */
    inv = invmod64_64(k_term,c_term,p); /* inv <-- -c/k (mod p) */
    PRE2_MULMOD64_INIT(inv);
    C16 = sc_lists[parity+1][0][0];
    for (j = 0, list = C16; (h = *list++) < SUBSEQ_MAX; j++)
      D64[j] = PRE2_MULMOD64(BJ64[SUBSEQ[h].d],inv,p); /* -c/(k*b^d) */
    PRE2_MULMOD64_FINI();
    m = steps[j].m;
    M = steps[j].M;
    return j;
  }
  else if (s > 0)
  {
    /* p = 1 (mod s), where s is not a power of 2. Check for r-th power
       residues for each prime power divisor r of s.
    */
    p_s = p/s;
  }
  else /* s < 0 */
  {
    /* p = 1 (mod 2^s), where s > 1. Check for r-th power residues for each
       divisor r of s. We handle this case seperately to avoid computing p/s
       using plain division.
    */
    p_s = p >> (-s);
#ifndef NDEBUG
    s = 1 << (-s);
#endif
  }

  /* For 0 <= r < s, X[r] <- 1/(b^r)^((p-1)/s) */
  X[0] = 1;
  X[1] = powmod64(inv,p_s,p);
  PRE2_MULMOD64_INIT(X[1]);
  for (r = 1; X[r] != 1; r++)
    X[r+1] = PRE2_MULMOD64(X[r],X[1],p);
  PRE2_MULMOD64_FINI();
  assert(s%r == 0);
  /* 1/(b^r)^((p-1)/s)=1 (mod p) therefore (1/(b^r)^((p-1)/s))^y=1 (mod p)
     for 0 <= y < s/r. (Could we do more with this?)
  */

  /* X[r] <- (-ck)^((p-1)/s) */
  X[r] = powmod64((c_term == -1) ? k_term : p-k_term, p_s, p);

  /* Find h such that X[h]=X[r], i.e. (-ckb^h)^((p-1)/r)=1 (mod p),
     or h=r if not found. */
  for (h = 0; X[h] != X[r]; h++)
    ;

  if (h < r && (C16 = sc_lists[parity+1][divisor_index[r]][h]) != NULL)
  {
    /* -ckb^n is an r-power residue for at least one term k*b^n+c of
       this sequence.
    */

    /* Precompute 1/b^d (mod p) for 0 <= d <= Q, as necessary. */
    bQ = climb_ladder(sc_ladders[parity+1][divisor_index[r]][h],inv,p);
    inv = invmod64_64(k_term,c_term,p); /* inv <-- -c/k (mod p) */
    PRE2_MULMOD64_INIT(inv);
    for (j = 0, list = C16; (h = *list++) < SUBSEQ_MAX; j++)
    {
      /* -ckb^d is an r-th power residue for at least one term
         (k*b^d)*(b^Q)^(n/Q)+c of this subsequence.
      */
      D64[j] = PRE2_MULMOD64(BJ64[SUBSEQ[h].d],inv,p); /* -c/(k*b^d) */
    }
    PRE2_MULMOD64_FINI();
    m = steps[j].m;
    M = steps[j].M;
    return j;
  }

  return 0;
}

#ifdef HAVE_vector
static uint32_t (*baby_steps)(uint64_t);
static void (*giant_steps)(uint64_t,uint32_t);
#else
#define baby_steps baby_steps_1
#define giant_steps giant_steps_1
#endif

static void bsgs64(uint64_t p, int parity)
{
  uint32_t i, j, k, cc;

  mod64_init(p);

  if ((cc = setup64(p,parity)) > 0)
  {
    /* Baby steps. */
    if ((i = baby_steps(p)) > 0)
    {
      /* i is the order of b (mod p). This is all the information we need to
         determine every solution for this p, so no giant steps are needed. */

      for (k = 0; k < cc; k++)
        for (j = lookup64(D64[k]); j < m*M; j += i)
          eliminate_term(C16[k],sieve_low+j,p);
    }
    else
    {
      /* First giant step. */
      for (k = 0; k < cc; k++)
        if ((j = lookup64(D64[k])) != HASH_NOT_FOUND)
          eliminate_term(C16[k],sieve_low+j,p);

      /* Remaining giant steps. */
      if (M > 1)
        giant_steps(p,cc);
    }
  }

  mod64_fini();
}

static void choose_steps(uint32_t r, uint32_t s)
{
  uint32_t m, M, max_elts;

  /*
    r = range of n, s = number of subsequences.
    In the worst case we will do do one table insertion and one mulmod
    for m baby steps, then s table lookups and s mulmods for M giant
    steps. The average case depends on how many solutions are found
    and how early in the loop they are found, which I don't know how
    to analyse. However for the worst case we just want to minimise
    m + s*M subject to m*M >= r, which is when m = sqrt(s*r).
  */

  M = MAX(1,rint(sqrt((double)r/s)));
  m = MIN(r,ceil((double)r/M));

#if HASHTABLE_SIZE_OPT
  max_elts = (hashtable_size) ? hashtable_size/sizeof(HASH_T) : HASH_MAX_ELTS;
#else
  max_elts = HASH_MAX_ELTS;
#endif

  if (m > max_elts)
  {
    /* There are three ways to handle this undesirable case:
       1. Allow m > HASH_MAX_ELTS (undersize hash table).
       2. Restrict m <= HASH_MAX_ELTS (suboptimal baby/giant-step ratio).
       3. Recompile with SHORT_HASHTABLE=0.
    */
    M = ceil((double)r/max_elts);
    m = ceil((double)r/M);
  }

  assert(m <= max_elts);

  steps[s].m = m;
  steps[s].M = M;
}

#ifdef HAVE_vector
/* Increase baby steps as far as the next multiple of vec_len if doing so
   will reduce the number of giant steps. This minimises the extra work due
   to doing mulmods in blocks of vec_len. Assumes vec_len is a power of 2.
*/
static void trim_steps(uint32_t r, int vec_len, uint32_t max_baby_steps)
{
  uint32_t i, m, M;

  for (i = 1; i <= subseq_count; i++)
  {
    m = steps[i].m;
    M = steps[i].M;
    if (m & (vec_len-1)) /* m is not a multiple of vec_len */
    {
      m = (m | (vec_len-1)) + 1; /* next multiple of vec_len. */
      while (M > 1 && m*(M-1) >= r) /* minimise M */
        M--;
      while ((m-1)*M >= r) /* minimise m */
        m--;
      if (m <= max_baby_steps)
      {
        steps[i].m = m;
        steps[i].M = M;
      }
    }
  }
}
#endif

#if PRE_POWER
static void pre_power_init(uint64_t pmin)
{
  uint32_t i;
  uint64_t bi;

  dQ = subseq_Q;
  bdQ = b_term;
  for (i = 2, bi = b_term; i <= subseq_Q && bi < pmin/b_term; i++)
  {
    bi *= b_term;
    if (subseq_Q % i == 0)
    {
      dQ = subseq_Q/i;
      bdQ = bi;
    }
  }
}
#endif

#ifdef HAVE_vector
/* Choose which mulmod scheme to use for bsgs.
 */
typedef struct {
  uint32_t (*fun)(uint64_t);
  const char *desc;
  int vec_len;
} baby_fun_t;

typedef struct {
  void (*fun)(uint64_t,uint32_t);
  const char *desc;
} giant_fun_t;

static const baby_fun_t baby_step_funs[] = {
#ifdef HAVE_vector_sse2
  { baby_steps_sse2_2, "sse2/2", 2 },
  { baby_steps_sse2_4, "sse2/4", 4 },
  { baby_steps_sse2_8, "sse2/8", 8 },
  { baby_steps_sse2_16, "sse2/16", 16 },
#endif
  { baby_steps_1, "gen/1", 1 },
  { baby_steps_2, "gen/2", 2 },
  { baby_steps_4, "gen/4", 4 },
  { baby_steps_8, "gen/8", 8 }
};

static const giant_fun_t giant_step_funs[] = {
#ifdef HAVE_vector_sse2
  { giant_steps_sse2_2, "sse2/2" },
  { giant_steps_sse2_4, "sse2/4" },
  { giant_steps_sse2_8, "sse2/8" },
  { giant_steps_sse2_16, "sse2/16" },
#endif
  { giant_steps_1, "gen/1" },
  { giant_steps_2, "gen/2" },
  { giant_steps_4, "gen/4" },
  { giant_steps_8, "gen/8" }
};

#define NUM_TEST_PRIMES 10
static const uint64_t test_primes[NUM_TEST_PRIMES] = {
  UINT64_C(2250000000000023),
  UINT64_C(2250000000000043),
  UINT64_C(2250000000000059),
  UINT64_C(2250000000000061),
  UINT64_C(2250000000000079),
  UINT64_C(2250000000000089),
  UINT64_C(2250000000000113),
  UINT64_C(2250000000000163),
  UINT64_C(2250000000000191),
  UINT64_C(2250000000000209) };

#define NUM_BABY_FUN sizeof(baby_step_funs)/sizeof(baby_fun_t)
static int choose_baby_method(int print)
{
  uint64_t t0, t1, best[NUM_BABY_FUN];
  uint32_t cc;
  int i, j;

#if PRE_POWER
  pre_power_init(test_primes[0]);
#endif

  /* How many subsequences pass the power residue tests? Probably 1/4 or
     less pass, but the speed of the bsgs routine is more important when a
     larger number pass since then BSGS accounts for a larger proportion of
     the total work done. Benchmark assuming 1/2 have passed.
  */
  cc = (subseq_count+1)/2;
  m = steps[cc].m;
  for (i = 0; i < NUM_BABY_FUN; i++)
  {
    best[i] = UINT64_MAX;
    for (j = 0; j < NUM_TEST_PRIMES; j++)
    {
      mod64_init(test_primes[j]);
      t0 = timestamp();
      baby_step_funs[i].fun(test_primes[j]);
      t1 = timestamp();
      best[i] = MIN(best[i],t1-t0);
      mod64_fini();
    }
  }

  for (i = 1, j = 0; i < NUM_BABY_FUN; i++)
    if (best[i] < best[j])
      j = i;

  if (print >= 2)
    for (i = 0; i < NUM_BABY_FUN; i++)
      report(1,"Best time for baby step method %s: %"PRIu64".",
             baby_step_funs[i].desc,best[i]);

  return j;
}

#define NUM_GIANT_FUN sizeof(giant_step_funs)/sizeof(giant_fun_t)
static int choose_giant_method(int print)
{
  uint64_t t0, t1, best[NUM_GIANT_FUN];
  uint32_t cc;
  int i, j;

#if PRE_POWER
  pre_power_init(test_primes[0]);
#endif

  cc = (subseq_count+1)/2;
  m = steps[cc].m;
  M = MAX(2,steps[cc].M);

  mod64_init(test_primes[0]);
  D64[0] = 23;
  for (i = 1; i < cc; i++)
    D64[i] = mulmod64(D64[i],D64[0],test_primes[0]);
  mod64_fini();

  for (i = 0; i < NUM_GIANT_FUN; i++)
  {
    best[i] = UINT64_MAX;
    for (j = 0; j < NUM_TEST_PRIMES; j++)
    {
      mod64_init(test_primes[j]);
      baby_steps(test_primes[j]);
      t0 = timestamp();
      giant_step_funs[i].fun(test_primes[j],cc);
      t1 = timestamp();
      best[i] = MIN(best[i],t1-t0);
      mod64_fini();
    }
  }

  for (i = 1, j = 0; i < NUM_GIANT_FUN; i++)
    if (best[i] < best[j])
      j = i;

  if (print >= 2)
    for (i = 0; i < NUM_GIANT_FUN; i++)
      report(1,"Best time for giant step method %s: %"PRIu64".",
             giant_step_funs[i].desc,best[i]);

  return j;
}

/* Set baby_steps, giant_steps functions. Return baby step vector size.
*/
static int choose_bsgs_methods(int print)
{
  benchmarking = 1; /* Don't check or report any factors. */

  int baby_method;
  if (baby_method_opt == NULL)
    baby_method = choose_baby_method(print);
  else
  {
    for (baby_method = 0; baby_method < NUM_BABY_FUN; baby_method++)
      if (strcmp(baby_method_opt,baby_step_funs[baby_method].desc)==0)
        break;
    if (baby_method >= NUM_BABY_FUN)
      error("Unknown baby step method `%s'.",baby_method_opt);
  }
  baby_steps = baby_step_funs[baby_method].fun;

  int giant_method;
  if (giant_method_opt == NULL)
    giant_method = choose_giant_method(print);
  else
  {
    for (giant_method = 0; giant_method < NUM_GIANT_FUN; giant_method++)
      if (strcmp(giant_method_opt,giant_step_funs[giant_method].desc)==0)
        break;
    if (giant_method >= NUM_GIANT_FUN)
      error("Unknown giant step method `%s'.",giant_method_opt);
  }
  giant_steps = giant_step_funs[giant_method].fun;

  if (print)
    report(1,"Baby step method %s, giant step method %s.",
          baby_step_funs[baby_method].desc,giant_step_funs[giant_method].desc);

  benchmarking = 0;

  return baby_step_funs[baby_method].vec_len;
}
#endif /* HAVE_vector */

static void init_bsgs(void)
{
  uint32_t r, s, sieve_high, max_baby_steps;

  sieve_low = n_min / subseq_Q;
  sieve_high = n_max / subseq_Q;

  r = sieve_high - sieve_low + 1;
  steps = xmalloc((subseq_count+1)*sizeof(steps_t));

  for (s = 1, max_baby_steps = 0; s <= subseq_count; s++)
  {
    choose_steps(r,s);
    assert(sieve_high < sieve_low+steps[s].m*steps[s].M);
    max_baby_steps = MAX(max_baby_steps,steps[s].m);
  }

  if (verbose_opt >= 2)
   report(1,"BSGS range: %"PRIu32"*%"PRIu32" - %"PRIu32"*%"PRIu32".",
          steps[1].m,steps[1].M,steps[subseq_count].m,steps[subseq_count].M);

  /* Allow room for vector operations to overrun the end of the array.
   */
  D64 = xmemalign(64,(subseq_count+7)*sizeof(uint64_t));
  memset(D64,0,(subseq_count+7)*sizeof(uint64_t));
  init_hashtable(max_baby_steps);

#ifdef HAVE_vector
  {
    int vec_len;
    choose_bsgs_methods(0); /* Dummy run to prep cache. */
    vec_len = choose_bsgs_methods(verbose_opt);
    trim_steps(r,vec_len,max_baby_steps);
  }
#endif /* HAVE_vector */

#if PRE_POWER
  pre_power_init(p_min);
#endif
}

static void fini_bsgs(void)
{
  fini_hashtable();
  xfreealign(D64);
  free(steps);
}

void CODE_PATH_NAME(sieve)(void)
{
  init_bsgs();
  init_prime_sieve(p_max);
  start_srsieve();
#if CHECK_FOR_GFN
  if (seq_gfn)
    prime_sieve_gfn(p_min,p_max,bsgs64,seq_gfn);
  else
    prime_sieve(p_min,p_max,bsgs64);
#else
  prime_sieve(p_min,p_max,bsgs64);
#endif
  finish_srsieve("range is complete",p_max);
  fini_prime_sieve();
  fini_bsgs();
}
